# -*- coding: utf-8 -*-
import time
from traceback import format_exc

from cerberus import Validator

from concurrency.FeedbackThread import FeedbackThread
from entity.FeedBack import pull_stream_feedback
from entity.TaskParam import Param
from enums.ExceptionEnum import ExceptionType
from enums.StatusEnum import ExecuteStatus, PushStreamStatus
from exception.CustomerException import ServiceException
from loguru import logger
from multiprocessing import Queue
from concurrency.IntelligentRecognitionProcess import PushStreamProcess
from util.CpuUtils import print_cpu_ex_status
from util.KafkaUtils import CustomerKafkaConsumer
from util.QueUtil import put_queue
from util.RWUtils import getConfigs

'''
   分发服务
'''


class DispatcherService:
    __slots__ = (
        '__base_dir',
        '__context',
        '__feedbackThread',
        '__listeningProcesses',
        '__fbQueue',
        '__topics',
        '__analysisType',
        '__resource_status'
    )

    """
        初始化
    """

    def __init__(self, base_dir, active):
        self.__context = getConfigs(base_dir, 'config/application.json')
        self.__base_dir = base_dir
        if len(active) > 0:
            self.__context["dsp"]["active"] = active[0]
        logger.info("当前环境为: {}", self.__context["dsp"]["active"])
        self.__resource_status = False
        self.__feedbackThread = None

        self.__listeningProcesses = {}
        self.__fbQueue = Queue()

        self.__topics = (self.__context["kafka"]["topic"]["dsp-push-stream-task-topic"],)
        self.__analysisType = {
            self.__topics[0]: ('9998', lambda x, y: self.pullStream(x, y),
                               lambda x, y, z, r: self.push_stream_method(x, y, z, r))
        }
        self.start_service()

    @staticmethod
    def check_pull_stream_msg(msg, schema):
        try:
            v = Validator(schema, allow_unknown=True)
            result = v.validate(msg)
            if not result:
                logger.error("参数校验异常: {}", v.errors)
                if msg.get("requestId"):
                    logger.error("参数校验异常: {}, requestId: {}", v.errors, msg.get("requestId"))
                    raise ServiceException(ExceptionType.ILLEGAL_PARAMETER_FORMAT.value[0],
                                           ExceptionType.ILLEGAL_PARAMETER_FORMAT.value[1])
            return result
        except ServiceException as s:
            raise s
        except Exception:
            logger.error("参数校验异常: {}", format_exc())
            raise ServiceException(ExceptionType.ILLEGAL_PARAMETER_FORMAT.value[0],
                                   ExceptionType.ILLEGAL_PARAMETER_FORMAT.value[1])

    def pullStream(self, message, analysisType):
        if "start" == message.get("command"):
            logger.info("开始推流, requestId:{}", message.get("requestId"))
            if self.__resource_status or len(self.__listeningProcesses) >= 10:
                raise ServiceException(ExceptionType.NO_RESOURCES.value[0],
                                       ExceptionType.NO_RESOURCES.value[1])
            self.startPushStreamProcess(message, analysisType)
        elif "stop" == message.get("command"):
            self.stopPushStreamProcess(message, analysisType)
        else:
            pass

    def startPushStreamProcess(self, msg, analysisType):
        check_result = self.check_pull_stream_msg(msg, START_SCHEMA)
        if not check_result:
            return
        if self.__listeningProcesses.get(msg.get("requestId")):
            logger.warning("重复任务，请稍后再试！requestId:{}", msg.get("requestId"))
            return
        param = Param(self.__fbQueue, msg, analysisType, self.__base_dir, self.__context)
        srp = PushStreamProcess(param)
        srp.start()
        self.__listeningProcesses[msg.get("requestId")] = srp

    # 结束录屏进程
    def stopPushStreamProcess(self, msg, analysisType):
        check_result = self.check_pull_stream_msg(msg, STOP_SCHEMA)
        if not check_result:
            return
        srp = self.__listeningProcesses.get(msg.get("requestId"))
        if srp is None:
            logger.warning("未查询到该任务，无法停止任务！requestId:{}", msg.get("requestId"))
            put_queue(self.__fbQueue, {
                "pull_stream": pull_stream_feedback(msg.get("requestId"),
                                                    ExecuteStatus.FAILED.value[0],
                                                    ExceptionType.NOT_REQUESTID_TASK_EXCEPTION.value[0],
                                                    ExceptionType.NOT_REQUESTID_TASK_EXCEPTION.value[1])},
                      msg.get("requestId"))
            return
        srp.sendEvent({"command": "stop", "videoIds": msg.get("videoIds", [])})

    def push_stream_method(self, topic, message, analysisType, requestId):
        try:
            self.__analysisType.get(topic)[1](message, analysisType)
        except ServiceException as s:
            logger.error("消息监听异常：{}, requestId: {}", s.msg, requestId)
            if message.get("requestId"):
                videoInfo = [{"id": url.get("id"), "status": PushStreamStatus.FAILED.value[0]} for url in
                             message.get("videoUrls", []) if url.get("id")]
                self.__fbQueue.put({
                    "pull_stream": pull_stream_feedback(requestId,
                                                        ExecuteStatus.FAILED.value[0],
                                                        s.code,
                                                        s.msg, videoInfo)})
        except Exception:
            logger.error("消息监听异常：{}, requestId: {}", format_exc(), requestId)
            if message.get("requestId"):
                videoInfo = [{"id": url.get("id"), "status": PushStreamStatus.FAILED.value[0]} for url in
                             message.get("videoUrls", []) if url.get("id")]
                self.__fbQueue.put({
                    "pull_stream": pull_stream_feedback(message.get("requestId"),
                                                        ExecuteStatus.FAILED.value[0],
                                                        ExceptionType.SERVICE_INNER_EXCEPTION.value[0],
                                                        ExceptionType.SERVICE_INNER_EXCEPTION.value[1], videoInfo)})

    # 服务调用启动方法
    def start_service(self):
        # 初始化kafka监听者
        customerKafkaConsumer = CustomerKafkaConsumer(self.__context, topics=self.__topics)
        logger.info("(♥◠‿◠)ﾉﾞ  DSP【工具服务】启动成功   ლ(´ڡ`ლ)ﾞ")
        # 循环消息处理
        start_time = time.time()
        while True:
            try:
                # 检查任务进程运行情况，去除活动的任务
                self.check_process_task()
                # 检测cpu是否有服务资源
                start_time = self.check_service_resource(start_time)
                self.start_feedback_thread()
                # 拉取kafka消息
                msg = customerKafkaConsumer.poll()
                if msg is not None and len(msg) > 0:
                    for k, v in msg.items():
                        for m in v:
                            message = m.value
                            customerKafkaConsumer.commit_offset(m)
                            requestId = self.getRequestId(message.get("requestId"))
                            logger.info("当前拉取到的消息, topic:{}, offset:{}, partition: {}, body: {}, requestId:{}",
                                        m.topic, m.offset, m.partition, message, requestId)
                            topic_method = self.__analysisType.get(m.topic)
                            topic_method[2](m.topic, message, topic_method[0], requestId)
                else:
                    time.sleep(1)
            except Exception:
                logger.exception("主线程异常：{}", format_exc())

    '''
        考虑到requestId为空的场景
    '''

    @staticmethod
    def getRequestId(request_id):
        if not request_id:
            return '1'
        return request_id

    def check_service_resource(self, start_time, requestId=None):
        if len(self.__listeningProcesses) > 0:
            # gpu_result = print_gpu_ex_status(requestId)
            cpu_result = print_cpu_ex_status(self.__base_dir, requestId)
            # if gpu_result or cpu_result:
            # 如果没有服务器资源
            if cpu_result:
                self.__resource_status = True
                return time.time()
            # if not gpu_result and not cpu_result and time.time() - start_time > 30:
            if not cpu_result and time.time() - start_time > 30:
                self.__resource_status = False
                return time.time()
        return start_time

    def check_process_task(self):
        for requestId in list(self.__listeningProcesses.keys()):
            if not self.__listeningProcesses[requestId].is_alive():
                del self.__listeningProcesses[requestId]

    def start_feedback_thread(self):
        if self.__feedbackThread is None:
            self.__feedbackThread = FeedbackThread(self.__fbQueue, self.__context)
            self.__feedbackThread.setDaemon(True)
            self.__feedbackThread.start()
        start_time = time.time()
        retry_count = 0
        while True:
            if self.__feedbackThread.is_alive():
                break
            retry_count += 1
            if retry_count > 8:
                self.__feedbackThread = None
                logger.error("反馈线程异常重试失败！！！！！！")
                break
            if time.time() - start_time <= 3:
                logger.error("反馈线程异常等待中")
                time.sleep(1)
                continue
            logger.error("反馈线程异常重启中")
            self.__feedbackThread = FeedbackThread(self.__fbQueue, self.__context)
            self.__feedbackThread.setDaemon(True)
            self.__feedbackThread.start()
            start_time = time.time()
            continue


START_SCHEMA = {
    "requestId": {
        'type': 'string',
        'required': True,
        'empty': False,
        'regex': r'^[a-zA-Z0-9]{1,36}$'
    },
    "command": {
        'type': 'string',
        'required': True,
        'allowed': ["start", "stop"]
    },
    "videoUrls": {
        'type': 'list',
        'required': True,
        'nullable': False,
        'schema': {
            'type': 'dict',
            'required': True,
            'schema': {
                "id": {
                    'type': 'string',
                    'required': True,
                    'empty': False,
                    'dependencies': "pullUrl",
                    'regex': r'^[a-zA-Z0-9]{1,255}$'
                },
                "pullUrl": {
                    'type': 'string',
                    'required': True,
                    'empty': False,
                    'dependencies': "pushUrl",
                    'regex': r'^(https|http|rtsp|rtmp|artc|webrtc|ws)://\w.+$'
                },
                "pushUrl": {
                    'type': 'string',
                    'required': True,
                    'empty': False,
                    'dependencies': "id",
                    'regex': r'^(https|http|rtsp|rtmp|artc|webrtc|ws)://\w.+$'
                }
            }
        }
    }
}
STOP_SCHEMA = {
    "requestId": {
        'type': 'string',
        'required': True,
        'empty': False,
        'regex': r'^[a-zA-Z0-9]{1,36}$'
    },
    "command": {
        'type': 'string',
        'required': True,
        'allowed': ["start", "stop"]
    },
    "videoIds": {
        'type': 'list',
        'required': False,
        'nullable': True,
        'schema': {
            'type': 'string',
            'required': True,
            'empty': False,
            'regex': r'^[a-zA-Z0-9]{1,255}$'
        }
    }
}
